from __future__ import print_function
"""
Supplementary or Auxiliary Functions for Snobfit

Most of them are the functions of the Rough MATLAB-Numpy Equivalents.
"""

import numpy

# Some constants
eps   = numpy.finfo('d').eps
inf   = numpy.inf
nan   = numpy.nan

# Some internal functions of np
norm  = numpy.linalg.norm
inv   = numpy.linalg.inv
triu  = numpy.triu
isnan = numpy.isnan


# ----------------------------------------------------------
def diag(arr):
    return numpy.diag(arr.flatten())

def within(low, x, high):
    return numpy.logical_and(low <= x, high >= x)


def feval(funcName, *args):
     return eval(funcName)(*args)


def find(cond_array):
    #cond_array should already have np conditional applied
    # tup = numpy.where(*args,**kw)
    # if len(tup)==0:
    #     return numpy.array([],'int')
    # else:
    #     return tup[0]
    return (numpy.transpose(numpy.nonzero(cond_array.flatten()))).astype(int)
    # return numpy.array([i for (i, val) in enumerate(a) if func(val)])


def toCol(v):
    """
    v is a vector
    """
    return numpy.matrix(v.reshape(v.size, 1))


def toRow( m ):
    """
    m is a  n x 1 matrix
    """
    n = m.shape[0]
    return numpy.array( numpy.asarray(m).reshape(1,n)[0] )
    # numpy.array( m.A.reshape(1,n)[0] )


def vector(n):
     """
     Return a vector of the given length.
     """
     return numpy.empty(n,'d')

def ivector(n=0):
    """
    Return a int vector of the given length.
    """
    return numpy.zeros(n,'int')


def sort(x):
    ind = numpy.argsort(x)
    return x[ind], ind


def std(x):
    """
    STD(X) normalizes by (N-1) where N is the sequence length.
    This makes STD(X).^2 the best unbiased estimate of the variance
    if X is a sample from a normal distribution.
    """
    return numpy.std(x,ddof=1)


def max_(x):
    """
    (Y,I) = _max(x) returns the index I of the maximum values in vector x.
    If the values along the first non-singleton dimension contain more
    than one maximal element, the index of the first one is returned.
    """
    if x.size <= 0:
        return (numpy.array([None]), numpy.array([None]))
    idx = numpy.argmax(x)
    return x[idx], idx


def min_(x):
    """
    [Y,I] = MIN(X) returns the index of the minimum value in vector I.
    If the values along the first non-singleton dimension contain more
    than one minimal element, the index of the first one is returned.
    """
    if x.size <= 0:
        return (numpy.array([None]), numpy.array([None]))
    idx = numpy.argmin(x)
    return x[idx], idx

def maximum_(a, b):
    return numpy.max(numpy.concatenate((a, b)))

def extend(f, n):
    f = numpy.append(f, numpy.zeros((f.shape[0], n)), axis=1)
    return f


def toInt(x):
    # Make sure it is vector
    if x.shape[0]==1:
        x = numpy.reshape( numpy.asarray(x), x.shape[1] )

    Ix =  numpy.array( x, 'int' )
    # 10000 is big enough??? Fix me later.

    Ix[ numpy.where(Ix<0) ] = 10000
    return Ix


def isEmpty(x):
    if len(x)==0: return True
    if x.shape[0]==0: return True
    if x.shape[1]==0: return True
    return False


def removeByInd(a, ind):
    n = len(a) - len(ind)
    ret = numpy.zeros( n, 'int' )
    k = 0
    for i in xrange(len(a)):
        if (i not in ind):
            ret[k] = a[i]
            k += 1
    return ret


def duplicate(v, n, axis=0):
    #return np matrix
    return numpy.mat( v ).repeat(n, axis=axis)


def dup(v, n, axis=0):
    # Return a np array
    if axis == 0:
        return numpy.array( [v] ).repeat(n, axis=axis)
    else:
        return numpy.array( v ).repeat(n, axis=axis)


def dot3(a,b,c):
    # Calculate a*(b*c)
    return numpy.dot(a, numpy.dot(b,c) )


def crdot(m, v):
    n = len(v)
    _sum = 0.0
    for i in xrange(n):
        _sum += m[i,0]*v[i]

    return _sum


#---------------------------------------------------------------
_randstate = numpy.random.RandomState(6)
def rand(*args):
    """
    For debugging purposes, select an initial state that can be
    reproduced with octave:

    %%%%%
    % https://stackoverflow.com/questions/13735096/python-vs-octave-random-generator/13876670#13876670
    function state = mtstate(seed)

    state = uint32(zeros(625,1));

    state(1) = uint32(seed);
    for i=1:623,
       tmp = uint64(1812433253)*uint64(bitxor(state(i),bitshift(state(i),-30)))+i;
       state(i+1) = uint32(bitand(tmp,uint64(intmax('uint32'))));
    end
    state(625) = 1;

    % initialize
    rand('state', mtstate(4));
    %%%%%

    Further, when asking for multi-dimensional matrices of random numbers,
    octave and numpy order things differently, so fix that, too.
    """

    if len(args) == 1:
        return _randstate.rand(*args)
    elif len(args) == 2:
        res = numpy.zeros(args)
        for i in range(args[1]):
            res[:,i] = _randstate.rand(args[0])
        return res
    else:
        # more dims not needed (yet); deal with it if we get there ...
        raise NotImplementedError("no implementation for %d-dim" % len(args))

    res = numpy.zeros(args)
    idim = 0
    for arg in args:
        res[:,idim] = _randstate.rand(arg)
        idim += 1
    return res

#---------------------------------------------------------------
def rsort(x, w=None, remove_dups=True):
    """
    Sort x in increasing order, remove multiple entries,
    and adapt weights w accordingly x and w must both be a row or a column
    default input weights are w=1

    If w==None, the weighted empirical cdf is computed at x
    dof = len(x) at in

    Warning: when you use this function, make sure x and w is row vector
    """
    if  w is None:
        w = numpy.ones(x.shape)

    ind = numpy.argsort(x)
    x = x[ind]
    w = w[ind]

    n = len(x)

    # Remove dubplicates if requested
    xnew = numpy.append(x[1:n], inf)
    if remove_dups:
        ind = numpy.nonzero(xnew != x)
    else:
        ind = numpy.nonzero(range(1, len(x)+1))
    nn  = len(ind)
    x = x[ind]

    ww = numpy.zeros(nn)
    if nn>1:
        ww[0] = sum(w[ind[0]+1])
        for i in range(nn-1):
            ww[i+1] = sum(w[ind[i:i+1]])

    # Get cumulative sum of weights
    cdfx = numpy.cumsum(ww)

    # Adjust for jumps and normalize
    if cdfx[nn-1] != 0.:
        cdfx = (cdfx-0.5*ww)/cdfx[nn-1]
    dof = n

    return x, ww, cdfx, dof



from __future__ import print_function
import copy, string

__all__ = ['optset']

STANDALONE = True

def optset(optin=None, **optkwds):
    """
 Set the options in SnobFit.

  optout = optset(name1=val1, name2=val2, ...)
           creates the options dictionary with the named options
           set to something other than the defaults

  optout = optset(optin, name1=val1, name2=val2, ...);
           modifies the dictionary <optin> and creates a new structure
           <optout>

  And now for the lists. The details of the options, the motiviation
  for the defaults, and suggestions for their use live in the manual

 Scalar Options:

  minfcall: minimum number of function calls before considering stopping
            default: (problem dimension)*5

  maxmp:    maximum number of model points requested for the local fit
            default: 2*(problem dimension)+6

  maxfail:  maximum number of consecutive failures before iteration stops
            default: 5

  verbose: provide verbose (debugging) output
           default: False
   """

    if not optin:
    # Start from default structure
        optout = _create_defaults()
    else:
        optout = copy.copy(optin)

    for name, val in optkwds.items():
        setattr(optout, name.lower(), val)

    return optout


#-----
def _create_defaults():
    """Default values"""

    class OptSet(object):
        __slots__ = {
            'minfcall' : None,          # minimum number of function calls
            #
            'maxmp' : None,             # maximum number of model points (default set in driver)
            #
            'maxfail' : 5,              # stop iterating after maxfail failures to improve
            #
            'verbose' : False,          # provide verbose output
        }

        def __init__(self):
            for key, value in self.__slots__.items():
                setattr(self, key, value)

    return OptSet()



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice: #
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# [iNCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# [iNCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function xl, xu, x, f, nsplit, small, near, d, np, t, inew, fnan, u, v = \
   snobupdt(xl, xu, x, f, nsplit, small, near, d, np, t, xnew, fnew, fnan, u, v, u1, v1, dx)

 Updates the box parameters when a set of new points and their function
 values are added; i.e. the boxes containing more than one point are
 split and the nearest neighbors are computed or updated.

 Input:
  xl, xu        rows contain lower and upper bounds of the old boxes
  x             rows contain the old points
  f             f[j] contains the function value at x[j], its
                variation and other parameters
  nsplit        nsplit[j,i] number of times box j has been split along
                the ith coordinate
  small         small[j] is an integer-valued logarithmic volume measure
                of box j
  near          near[j] is a vector pointing to the nearest
                neighbors of x[j]
  d             d[j] is the maximal distance between x[j] and one of
                its neighbors
  np            np[j] is the number of times the function value of
                x[j] has been measured
  t             t[j] is np[j] times the variance of the function values
                measured for the point x[j]
  xnew          rows contain new points
  fnew          new function values and their variations
                fnew[j,0] = f(xnew[j]), fnew[j,1] = df(xnew[j])
  fnan          pointer to all old points where the function value could
                not be obtained
  u, v          box bounds
  u1, v1        box in which the new points are to be generated
  dx            resolution vector

 Output:
  xl, xu       updated version of xl, xu [including new boxes)
  x            updated version of x
  f            updated version of f
  nsplit       updated version of nsplit
  small        updated version of small
  near         updated version of near
  d            updated version of d
  np           updated version of np
  t            updated version of t
  inew         pointer pointing to the new boxes and boxes whose
               nearest neighbors have changed
  fnan         possibly updated version of fnan [if a function value
               was found for a point in the new iteration)
  u, v         possibly updated box bounds such that all new points
               are in the box
"""

from ._gen_utils import find, min_, rsort, sort
from ._snobinput import snobinput
from ._snobnn    import snobnn
from ._snobsplit import snobsplit
import numpy


def snobupdt(xl, xu, x, f, nsplit, small, near, d, np, t, xnew, fnew, fnan, u, v, u1, v1, dx):
    n = u.shape[1]      # dimension of the problem
    nneigh = n+5
    nxold = len(x)      # number of points from the previous iteration
    nxnew = len(xnew)

    inew = numpy.array([], dtype=int)
    if x.size > 0:
      # if any of the new points are already among the old points, they are
      # thrown away and the function value and its uncertainty are updated
        dismiss = []
        for j in range(nxnew):
            i = find(sum(abs(numpy.ones((nxold,1))*xnew[j]-x),1) == 0)
            if i.size > 0:
                if not find(fnan==i).any() and numpy.isfinite(f[i,1]): # point i had finite
                                                                       # function value
                    if not numpy.isnan(fnew[j,0]):
                        np[i] = np[i] + 1
                        delta = fnew[j,0] - f[i,0]
                        f[i,0] = f[i,0] + delta/np[i]
                        t[i] = t[i] + delta*(fnew[j,0]-f[i,0])
                        f[i,1] = math.sqrt(f[i,1]**2 + (delta*(fnew[j,0] - f[i,0]) \
                            + fnew[j,1]**2 - f[i,1]**2) / np[i])
                        inew = numpy.concatenate((inew, i))

                    dismiss = numpy.concatenate((dismiss, [j]))
                else: # point i had NaN function value
                    if not numpy.isnan(fnew[j,0]).any():
                        f[i,0] = fnew[j,0]
                        inew = numpy.concatenate((inew, i))
                        ii = find(fnan==i)
                        if ii.size > 0:
                            fnan[ii] = numpy.array([], dtype=int)

                    dismiss = numpy.concatenate((dismiss, [j]))

        xnew = numpy.delete(xnew, dismiss, 0)
        fnew = numpy.delete(fnew, dismiss, 0)
        nxnew = len(xnew)
        if not nxnew:
            inew = numpy.sort(inew)
            return xl, xu, x, f, nsplit, small, near, d, np, t, inew.astype(int), fnan, u, v

    xnew, fnew, npnew, tnew = snobinput(xnew, fnew)
    nxnew = xnew.shape[0]

    nx = nxold + nxnew	# current number of points

    if numpy.sum(numpy.vstack((xnew, u)).min(0) < u) or \
        numpy.sum(numpy.vstack((xnew, v)).max(0) > v) or \
        (numpy.minimum(u, u1) < u).any() or (numpy.maximum(v, v1) > v).any():
        xl, xu, small, u, v = snobnewb(xnew, xl, xu, small, u, v, u1, v1)

    if x.size > 0:
        x = numpy.concatenate((x, xnew))
    else:
        x = xnew.copy()
    inew = numpy.concatenate((inew, numpy.arange(nxold, nx)), 0)
    if f.size <= 0:
        f = fnew.copy()
    else:
        if fnew.shape[1] < f.shape[1]:
            fnew = numpy.append(fnew, numpy.zeros((len(fnew), f.shape[1]-fnew.shape[1])), axis=1)
        f = numpy.vstack((f, fnew))
    if np.size > 0:
        np = numpy.append(np, npnew)
    else:
        np = npnew.copy()
    if t.size > 0:
        t = numpy.append(t, tnew)
    else:
        t = tnew.copy()
    if not nxold:
        xl, xu, x, f, nsplit, small = snobsplit(x, f, u, v, None, u, v)
    else:
        par = numpy.zeros((nxnew,))
        for j in range(nxnew):
            xx = numpy.ones((nxold,1))*xnew[j]
            ind = find(numpy.sum( \
                numpy.logical_and(numpy.less_equal(xl, xx), numpy.less_equal(xx, xu)), 1) == n)
            if ind.size > 0:
                minsmall, ismall = min_(small[ind])
                par[j] = ind[ismall]

        par1, ww, cdfx, dof = rsort(par)
        inew = numpy.concatenate((inew, par1), 0)
        for j in par1.astype(int):
            ind = find(par==j)
            ind = ind + nxold
            spl = numpy.append([j], ind.flatten())
            xl0, xu0, x0, f0, nsplit0, small0 = \
                snobsplit(x[spl], f[spl], xl[j], xu[j], nsplit[j], u, v)
            nxj = len(ind) + 1    # number of points in box [xl[j],xu[j]]
            k = find(numpy.sum(x0 == numpy.ones((nxj,1))*x[j,:], axis=1) == n)
            if len(k) != 1:
                if 1 < len(k): k = k[min_(small[k])[1]]
                else: k = 0
            xl[j] = xl0[k]
            xu[j] = xu0[k]
            nsplit[j] = nsplit0[k]
            small[j] = small0[k]
            for k in range(nxj-1):
                k1 = ind[k]
                k2 = find(numpy.sum(x0 == numpy.ones((nxj,1))*x[k1,:],axis=1) == n)
                if len(k2) != 1:
                    if 1 < len(k2): k2 = k2[min_(small[k2])[1]]
                    else: k2 = 0
                ik1 = int(k1)
                if len(xl) <= ik1:
                    xl = numpy.append(xl, numpy.zeros((ik1-len(xl)+1, xl.shape[1])), axis=0)
                if len(xu) <= ik1:
                    xu = numpy.append(xu, numpy.zeros((ik1-len(xu)+1, xu.shape[1])), axis=0)
                if len(nsplit) <= ik1:
                    nsplit = numpy.append(nsplit, numpy.zeros((ik1-len(nsplit)+1, nsplit.shape[1])), axis=0)
                if len(small) <= ik1:
                    small = numpy.append(small, numpy.zeros((ik1-len(small)+1,)), axis=0)
                xl[k1] = xl0[k2]
                xu[k1] = xu0[k2]
                nsplit[k1] = nsplit0[k2]
                small[k1] = small0[k2]

    notnan = numpy.arange(0, nx)
    notnan = numpy.delete(notnan, find(numpy.isnan(f[:,0])))

    if notnan.size > 0:
        fmn = numpy.min(f[notnan,0])
        fmx = numpy.max(f[notnan,0])
    else:
        fmn = 1
        fmx = 0

    if nx >= nneigh+1 and fmn < fmx:
        if near.size <= 0:
            near = numpy.zeros((nx, nneigh), dtype=int)
        if d.size <= 0:
            d = numpy.zeros((nx,))

        for j in range(nxold, nx):
            jnear, jd = snobnn(x[j], x, nneigh, dx)
            if len(near) <= j:
                near = numpy.append(near, numpy.zeros((j-len(near)+1, nneigh), dtype=int), axis=0)
            near[j] = jnear

            if len(d) <= j:
                d = numpy.append(d, numpy.zeros((j-len(d)+1,)))
            d[j] = jd

        for j in range(nxold):
            if numpy.min(numpy.sqrt(numpy.sum((numpy.ones((nxnew,1))*x[j]-xnew)**2,1))) < d[j]:
                jnear, jd = snobnn(x[j], x, nneigh, dx)
                if len(near) <= j:
                    near = numpy.append(near, numpy.zeros((j-len(near)+1, nneigh)), axis=0)
                near[j] = jnear

                if len(d) <= j:
                    d = numpy.append(d, numpy.zeros((j-len(d)+1,)))
                d[j] = jd
                inew = numpy.concatenate((inew, [j]))

        inew = sort(inew)[0]
        d = d.reshape((1, len(d)))
    else:
        near = numpy.array([])
        d = numpy.inf*numpy.ones((1, nx))

    return xl, xu, x, f, nsplit, small, near, d, np, t, inew.astype(int), fnan, u, v


def snobnewb(xnew, xl, xu, small, u, v, u1, v1):
    nx = len(xl)
    n = nx and xl.shape[1] or 0
    uold = u
    vold = v
    u = numpy.concatenate((xnew, u)).min(0)
    v = numpy.concatenate((xnew, v)).max(0)
    u = numpy.min(u, u1)
    v = numpy.max(v, v1)
    i1 = find(u < uold)
    i2 = find(v > vold)
    ind = numpy.array([])
    for j in range(len(i1)):
        j1 = find(xl[:,i1[j]] == uold[i1[j]])
        ind = numpy.concatenate((ind, j1))
        xl[j1,i1[j]] = u[i1[j]]

    for j in range(len(i2)):
        j2 = find(xu[:, i2[j]] == vold[i2[j]])
        ind = numpy.concatenate((ind, j2))
        xu[j2,i2[j]] = v[i2[j]]

    if len(i1) + len(i2):    # at least one of the bounds was changed
        small = -numpy.sum(numpy.round(numpy.log2((xu-xl)/(numpy.ones((nx,1))*(v-u)))), axis=1)

    return xl, xu, small, u, v



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function y, f1 = snobqfit(j, x, f, near, dx, u, v)

 A quadratic model around the best point is fitted and minimized with
 minq over a trust region.
 
 Input:
  j            index of the best point
  x            the rows contain the points where the function has been
               evaluated
  f            corresponding function values, i.e., f[i] = f[x(i]]
  near         near[i] is a vector containing the indices of the nearest
               neighbors of the point x[i]
  dx           resolution vector, i.e. the ith coordinate of a point to be
               generated is an integer-valued multiple of dx[i]
  u, v         the points are to be generated in [u, v]
 
 Output:
  y            minimizer of the quadratic model around the best point
  f1           its estimated function value
"""

from ._gen_utils import rand, sort
from ._snoblocf import snobround
from .minq import minq
import numpy


def snobqfit(j, x, f, near, dx, u, v):
    n = x.shape[1]	# dimension of the problem
    K = min(len(x)-1, n*(n+3))
    nneigh = near.shape[1]
    x0 = x[j]
    f0 = f[j]
    distance = numpy.sum((x - numpy.outer(numpy.ones(len(x)), x0))**2, 1)
    dd, ind = sort(distance)
    ind = ind[1:K+1]
    nind = len(ind)
    d = numpy.abs(x[near[j]]-numpy.outer(numpy.ones(nneigh), x0)).max(0)
    d = numpy.maximum(d, dx)
    S = x[ind] - numpy.outer(numpy.ones(K), x0)
    R = numpy.triu(numpy.linalg.qr(S, 'r'))
    R = R[:n]      # unnecessary?
    L = numpy.linalg.inv(R).T
    sc = numpy.sum((S.dot(L.T))**2, 1)**(3/2.0)
    b = (f[ind]-f0)/sc
    A = numpy.concatenate((x[ind,:] - numpy.outer(numpy.ones(K), x0),
                        0.5*(x[ind,:] - numpy.outer(numpy.ones(K), x0))**2), 1)

    for i in range(n-1):
        B = (x[ind,i]-x0[i]).reshape(nind,1).dot(numpy.ones((1,n-i-1)))
        A = numpy.concatenate((A, B*(x[ind, i+1:n]-(numpy.ones((K, 1)).dot(x0[i+1:n][None,:])))), axis=1)

    A = A/(sc.reshape(nind,1).dot(numpy.ones((1, A.shape[1]))))
    y = numpy.linalg.lstsq(A, b, rcond=None)[0]

    G = numpy.zeros((n, n))
    for i in range(n):
        G[i,i] = y[n+i]

    l = 2*n
    for i in range(n-1):
        for j in range(i+1,n):
            G[i,j] = y[l]
            G[j,i] = y[l]
            l += 1

    g = y[:n]
    c = g - G.dot(x0.T).reshape(u.shape)
    dp = d.reshape(u.shape); x0p = x0.reshape(u.shape)
    y, f1, ierr, nsub = minq(f0-x0.dot(g) + 0.5*x0.dot(G.dot(x0.T)), c, G,
            numpy.maximum(x0p.T - dp.T, u.T), numpy.minimum(x0p.T + dp.T, v.T), 0)
    y = snobround(y.T, u, v, dx)
    nc = 0
    while (numpy.min(numpy.max(numpy.abs(x-numpy.outer(numpy.ones(len(x)), y))
            - numpy.outer(numpy.ones(len(x)), dx), axis=1)) < -numpy.spacing(1)) and nc < 10:
        u1 = numpy.maximum(x0 - d, u)
        v1 = numpy.minimum(x0 + d, v)
        y = u1 + rand(1,n)*(v1-u1)
        y = snobround(y, u, v, dx)
        nc += 1

    f1 = f0+(y-x0).dot(g) + 0.5*((y-x0).dot(G.dot((y-x0).T)))

    return y.flatten(), float(f1)



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobinput.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function x, f, np, t = snobinput(x, f)

 Checks whether there are any duplicates among the points given by the rows of
 x, throws away the duplicates and computes their average function values and an
 estimated uncertainty.

 Input:
  x            numpy array of points
  f            numpy array, f[j,0] is the function value of x[j] and f[j,1] is its
               uncertainty

 Output: (x, f, np, t)
  x            updated version of x (possibly some points have been deleted)
  f            updated version of f (f[j,0] is the average function value and
               f[j,1] the estimated uncertainty pertaining to x[j])
  np           np[j] is the number of times the row x[j] appeared in the
               input version of x
  t - t[j] is np[j] times the variance of the function values measured
               for point x[j]
"""

import math, numpy


def snobinput(x, f):
    if x.size <= 0:
        return x, f, numpy.array([]), numpy.array([])
    sx = len(x)                   # num points
    n = sx and x.shape[1] or 0    # num parameters per point
    i = 0

    nullp = numpy.ones(sx)
    t = numpy.ones(sx)

    while i < sx:
        j = i + 1
        ind = numpy.array([])

        while j < sx:
            if numpy.array_equal(x[i],x[j]):
                ind = numpy.concatenate((ind,[j]))
            j += 1

        if (len(ind) != 0):
            ind = numpy.concatenate(([i],ind)).astype(int)
            ind1 = numpy.transpose(numpy.nonzero(numpy.isnan(f[ind,0]))).astype(int)

            if len(ind1) < len(ind):
                ind = numpy.delete(ind, ind1).astype(int)
                nullp[i] = len(ind)

                fbar = numpy.sum(f[ind,0])/nullp[i]
                t[i] = numpy.sum((f[ind,0]-fbar)**2)
                f[i,0] = fbar
                f_dev = f[i,1]
                if(not isinstance(f[i,1], numpy.float64) ):
                    f_dev = sum(f[i,1])

                f[i,1] = math.sqrt((f_dev**2+t[i])/nullp[i])
            else:
                nullp[i] = 1
                t[i] = 0

            x = numpy.delete(x, ind[1:], 0)
            f = numpy.delete(f, ind[1:], 0)
            sx = x.shape[0]
        else:
            nullp[i] = 1
            t[i] = 0

        i += 1

    return x, f, nullp, t



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function y = snobpoint(x, xl, xu, f0, g, sigma, u, v, dx)

 For a box [xl, xu] containing a point x, a point y in the intersection
 of [xl, xu] and [u, v] is constructed such that it is both not close to
 x and to the boundary of [xl, xu] and its function value is estimated
 from a local quadratic model around x.
 
 Input:
  x            point contained in [xl, xu]
  xl, xu       box bounds
  u, v         the point is to be generated in [u, v]
  f0           f0[1] is the function value at x, f0[2] is its uncertainty
  g, G, sigma  the local quadratic model around x is given by
               q(y) = f0[0]+g*(y-x)'+sigma*((y-x)*diag(D)*(y-x)'+f0[1])
               for a row vector y, where D = f0[1]/dx**2
  dx           resolution vector
 
 Output:
  y            point in the intersection of [xl, xu] and [u, v]
  f            corresponding estimated function value
"""

from ._snoblocf  import snobround
import numpy


def snobpoint(x, xl, xu, f0, g, sigma, u, v, dx):
    n = len(x)
    y = numpy.zeros(n)
    for i in range(n):
        if x[i] - xl[i] > xu[i] - x[i]:
            y[i] = 0.5*(xl[i] + x[i])
        else:
            y[i] = 0.5*(x[i] + xu[i])

    y = numpy.minimum(numpy.maximum(y, u), v)
    y = snobround(y, numpy.maximum(xl,u), numpy.minimum(xu,v), dx)
    D = f0[1]/dx**2
    f = f0[0] + g.dot((y-x).T) + sigma*((y-x).dot(numpy.diag(D).dot((y-x).T))+f0[1])
    return y, float(f)



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
  function x, fct, ier, nsub = minq(gam, c, G, xu, xo, prt, xx)

  Minimizes an affine quadratic form subject to simple bounds:
  using coordinate searches and reduced subspace minimizations
  using LDL^T factorization updates
     min    fct =  gam + c^T x + 0.5 x^T G x
     s.t.   x in [xu:xo]    # xu< = xo is assumed
  where G is symmetric n x n
  (if G is indefinite: only a local minimum is found).

  Output:
  prt           print level
  xx            initial guess (optional)
  x             minimizer (but unbounded direction if ier = 1)
  fct           optimal function value
  ier            0 (local minimizer found)
                 1 (unbounded below)
                99 (maxit exceeded)
"""

from ._gen_utils import diag, find
from .minq_subroutines import getalp, ldldown, ldlup
import copy, logging, math, numpy

logger = logging.getLogger('SKQ.SnobFit.minq')


def minq(gam, c, G, xu, xo, prt, xx=None):
    prt = 0

    c = c.reshape(xu.shape)
    convex = 0
    n = len(G)
    hpeps = 100 * numpy.spacing(1)  # perturbation in last two digits
    maxit = 3 * n      # maximal number of iterations
                       # this limits the work to about 1+4*maxit/n matrix multiplies
                       # usually at most 2*n iterations are needed for convergence
    nitrefmax = 3      # maximal number of iterative refinement steps

  # initialize trial point xx: function value fct and gradient g
    if xx is None:
      # cold start with absolutely smallest feasible point
        xx = numpy.zeros(n).reshape(xu.shape)

  # force starting point into the box
    xx = numpy.maximum(xu, numpy.minimum(xx, xo))

  # initialize factorization
    K = numpy.zeros(n, dtype=int) # initially no rows in factorization
    L = numpy.eye(n)
    dd = numpy.ones(n)            # LDL^T factorization of G_KK
    #if issparse('G'):
    #    print('better: use symbolic factorization (not yet implemented)')

  # dummy initialization of indicator of free variables
  # will become correct after first coordinate search
    free = numpy.zeros(n, dtype=int)
    nfree = 0
    nfree_old = -1

    fct = numpy.inf               # best function value
    nsub = 0                      # number of subspace steps
    unfix = 1                     # allow variables to be freed in csearch?
    nitref = 0                    # no iterative refinement steps so far
    improvement = True            # improvement expected

  ########################################################################
  # main loop: alternating coordinate and subspace searches
    while True:
        logger.debug('enter main loop')

        if numpy.linalg.norm(xx, numpy.inf) == numpy.inf:
            error('infinite xx in minq.m')

        g = G.dot(xx) + c
        fctnew = float(gam + 0.5*xx.T.dot(c+g))
        if not improvement:
          # good termination
            logger.debug('terminate: no improvement in coordinate search')
            ier = 0
            break

        elif nitref > nitrefmax:
          # good termination
            logger.debug('terminate: nitref>nitrefmax')
            ier = 0
            break

        elif nitref > 0 and nfree_old == nfree and fctnew >= fct:
          # good termination
            logger.debug('terminate: nitref > 0 and nfree_old == nfree and fctnew >= fct')

            ier = 0
            break

        elif nitref == 0:
            x = xx[:]
            fct = min(fct, fctnew)
            logger.debug('fct: %s', fct)

        else:  # more accurate g and hence f if nitref>0
            x = xx[:]
            fct = fctnew
            logger.debug('fct: %s', fct)

        if nitref == 0 and nsub >= maxit:
            logger.debug('incomplete minimization (too many iterations): increase maxit')
            logger.info('iteration limit exceeded')
            ier = 99
            break

        ######################################################################
        # coordinate search
        count = 0        # number of consecutive free steps
        k = -1           # current coordinate searched
        while True:
            while count <= n:
              # find next free index (or next index if unfix)
                count += 1
                if k == n-1:
                    k = -1
                k += 1
                if free[k] or unfix:
                    break

            if count > n:
              # complete sweep performed without fixing a new active bound
                break

            q = G[:,k]
            alpu = xu[k] - x[k]
            alpo = xo[k] - x[k]      # bounds on step

          # find step size
            alp, lba, uba, ier = getalp(alpu, alpo, float(g[k]), float(q[k]))
            if ier:
                x = numpy.zeros(n)
                if lba:
                    x[k] = -1
                else:
                    x[k] = 1

                if logger.getEffectiveLevel() >= logging.DEBUG:
                    logger.debug('function unbounded below, unbounded direction returned')
                    logger.debug('possibly caused by roundoff')
                    logger.debug('f(alp*x) = gam+gam1*alp+gam2*alp^2/2: where')
                    logger.debug('gam1 = %s', (c.T).dot(x))
                    logger.debug('gam2 = %s', (x.T).dot(G.dot(x)))
                    logger.debug('ddd = %s', diag(G))
                    logger.debug('min_diag_G = %s', ddd.min())
                    logger.debug('max_diag_G = %s', ddd.max())

                return x, fct, ier, nsub       

            xnew = x[k] + alp
            if nitref > 0:
                logger.debug('xnew: %s alp: %s', xnew, alp)

            if lba or xnew <= xu[k]:
              # lower bound active
                logger.debug('%d at lower bound', k)
                if alpu != 0:
                    x[k] = xu[k]
                    g = g + (alpu*q).reshape(g.shape)
                    count = 0
                free[k] = 0
            elif uba or xnew >= xo[k]:
                # upper bound active
                logger.debug( '%d at upper bound', k)
                if alpo != 0:
                    x[k] = xo[k]
                    g = g + (alpo*q).reshape(g.shape)
                    count = 0
                free[k] = 0
            else:
                # no bound active
                logger.debug('%d free', k)
                if numpy.spacing(1) < abs(alp):
                    if not free[k]:
                        logger.debug('unfixstep: %s %s', x[k], alp)

                    x[k] = xnew
                    g = g + (alp*q).reshape(g.shape)
                    free[k] = 1

            # end of coordinate search

        ######################################################################
        nfree = int(numpy.sum(free))
        if unfix and nfree_old == nfree:
          # in exact arithmetic: we are already optimal
          # recompute gradient for iterative refinement
            g = G.dot(x) + c
            nitref += 1
            logger.debug('optimum found; iterative refinement tried')

        else:
            nitref = 0

        nfree_old = nfree
        g = g.reshape(c.shape)
        gain_cs = float(fct - gam -0.5*x.T.dot(c+g))
        improvement = gain_cs > 10*numpy.spacing(1) or not unfix

      # print (0,1) profile of free and return the number of nonnp.zeros
        #nfree = pr01('csrch ', free)
        logger.debug('gain_cs: %s', gain_cs)

      # subspace search
        xx = x[:]
        if not improvement or nitref >= nitrefmax:
          # optimal point found or enough refinement steps - nothing done
            pass
        elif nfree == 0:
            pass
          # no free variables - no subspace step taken
            logger.debug('no free variables - no subspace step taken')
            unfix = 1
        else:
          # take a subspace step
            nsub += 1

            fct_cs = gam+0.5*x.T.dot(c+g.reshape(c.shape))
            format='*** nsub = %4.0f fct = %15.6e fct_cs = %15.6e'
            logger.debug(format % (nsub, fct, fct_cs))

          # downdate factorization
            for j in find(free < K):    # list of newly active indices
                L, dd = ldldown(L, dd, int(j))
                K[j] = 0
                if logger.getEffectiveLevel() >= logging.DEBUG:
                    logger.debug('downdate; fact_ind: %s', find(K).flatten())

          # update factorization or find indefinite search direction
            definite = 1
            for j in find(free > K):   # list of newly freed indices
              # later: speed up the following by passing K to ldlup!
                p = numpy.zeros(n)
                if n > 1:
                    idk = find(K)
                    for kk in idk:
                        p[kk] = G[kk, j]
                p[j] = G[j, j]
                L, dd, p = ldlup(L, dd, int(j), p)
                definite = p.size <= 0
                if not definite:
                    logger.debug('indefinite or illconditioned step')
                    break
                K[j] = 1
                if logger.getEffectiveLevel() >= logging.DEBUG:
                    logger.debug('update; fact_ind %s', find(K).flatten())

            if definite:
              # find reduced Newton direction
                p = numpy.zeros(n)
                idk = find(K).flatten()
                for kk in idk:
                    p[kk] = g[kk]
                p = -numpy.linalg.solve(L.T, numpy.linalg.solve(L, p)/dd)
                # p will remain zero and ignored
                if logger.getEffectiveLevel() >= logging.DEBUG:
                    logger.debug('reduced Newton step; fact_ind: %s', find(K).flatten())

          # set tiny entries to zero
          # p = (x+p)-x
            for ii in range(len(p)):
                if abs(p[ii]) < 100*numpy.spacing(1):
                    p[ii] = 0.0
            ind = find(p != 0).flatten()
            if ind.size <= 0:
              # zero direction
                logger.debug('zero direction')
                unfix = 1
                continue

          # find range of step sizes
            pp = p[ind].reshape(x[ind].shape)
            oo = (xo[ind]-x[ind])/pp
            uu = (xu[ind]-x[ind])/pp
            alpu = -numpy.inf 
            if oo[pp<0].size > 0: alpu = max(alpu, numpy.max(oo[pp<0]))
            if uu[pp>0].size > 0: alpu = max(alpu, numpy.max(uu[pp>0]))
            alpo =  numpy.inf
            if oo[pp>0].size > 0: alpo = min(alpo, numpy.min(oo[pp>0]))
            if uu[pp<0].size > 0: alpo = min(alpo, numpy.min(uu[pp<0]))
          # TODO: original had <= and =>, and alpo == 0.0 happens
            if alpo < 0 or alpu > 0:
                logger.debug("current alpo, alpu: %f, %f", alpo, alpu)
                raise RuntimeError('programming error: no alp')

          # find step size
            gTp = g.T.dot(p)
            agTp = numpy.abs(g).T.dot(numpy.abs(p))
            if abs(gTp) < 100*numpy.spacing(1)*agTp:
              # linear term consists of roundoff only
                gTp = 0
            pTGp = p.T.dot(G.dot(p))
            if convex: pTGp = max(0, pTGp)
            if not definite and pTGp > 0:
                logger.debug('tiny pTGp = %s set to zero', pTGp)
                pTGp = 0

            alp, lba, uba, ier = getalp(alpu, alpo, gTp, pTGp)
            if ier:
                x = numpy.zeros(n)
                if lba: x = -p
                else: x = p
                if logger.getEffectiveLevel() >= logging.DEBUG:
                    qg = gTp/agTp
                    qG = pTGp/(numpy.linalg.norm(p, 1)**2*numpy.linalg.norm(G[:], numpy.inf))
                    lam = numpy.linalg.eig(G)
                    lam1 = numpy.min(lam)/numpy.max(abs(lam))
                    logger.debug("minq: function unbounded below\n"
                                 "  unbounded subspace direction returned\n"
                                 "  possibly caused by roundoff\n"
                                 "  regularize G to avoid this!")

                    logger.debug('f(alp*x)=gam+gam1*alp+gam2*alp^2/2, where')
                    logger.debug('gam1 = %s', c.T.dot(x))
                    logger.debug('rel1 = %s', gam1/(numpy.abs(c).T.dot(numpy.abs(x))))
                    gam2 = x.T.dot(G.dot(x))
                    if convex: gam2 = max(0, gam2)
                    logger.debug('gam2 = %s', gam2)
                    logger.debug('rel2 = %s', gam2/(numpy.abs(x).T.dot((numpy.abs(G).dot(numpy.abs(x))))))
                    logger.debug('ddd = %s', diag(G))
                    logger.debug('min_diag_G = %s', min(ddd))
                    logger.debug('max_diag_G = %s', max(ddd))

                return x, fct, ier, nsub

            unfix = not (lba or uba)   # allow variables to be freed in csearch?
            
          # update of xx
            for k in range(len(ind)):
              # avoid roundoff for active bounds
                ik = int(ind[k])
                if alp == uu[k]:
                    xx[ik] = xu[ik]
                    free[ik] = 0
                elif alp == oo[k]:
                    xx[ik] = xo[ik]
                    free[ik] = 0
                else:
                    logger.debug('generic update')
                    xx[ik] = xx[ik]+alp*p[ik]

                if numpy.abs(xx[ik]) == numpy.inf:
                    logger.debug("%s %s %s", ik, alp, p[ik])
                    raise RuntimeError('infinite xx in minq.py')

            nfree = sum(free)

      # print (0:1) profile of free and return the number of nonzeros
        #nfree = pr01('ssrch ', free)
        if unfix and numpy.sum(nfree) < n:
            logger.debug('bounds may be freed in next csearch')


  # end of main loop
    logger.debug('fct: %s', fct)

    return x, fct, ier, nsub



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ._gen_utils import diag, find
import copy, numpy, logging

logger = logging.getLogger('SKQ.SnobFit.minq')


#-----
def getalp(alpu, alpo, gTp, pTGp):
    """
     function alp, lba, uba, ier = getalp(alpu, alpo, gTp, pTGp)

     Get minimizer alp in [alpu, alpo] for a univariate quadratic
            q(alp) = alp*gTp+0.5*alp^2*pTGp

     lba       lower bound active
     uba       upper bound active

     ier       0 (finite minimizer)
               1 (unbounded minimum)
    """

    lba = False
    uba = False

    # determine unboundedness
    ier = False
    if alpu == -numpy.inf and (pTGp < 0 or (pTGp == 0 and gTp > 0)):
        ier = True; lba = True

    if alpo ==  numpy.inf and (pTGp < 0 or (pTGp == 0 and gTp < 0)):
        ier = True; uba = True

    if ier:
        return numpy.nan, lba, uba, ier

    # determine activity
    if pTGp == 0 and gTp == 0:
        alp = 0
    elif pTGp <= 0:
        # concave case minimal at a bound
        if alpu == -numpy.inf:
            lba = False
        elif alpo == numpy.inf:
            lba = True
        else:
            lba = bool((2*gTp + (alpu+alpo)*pTGp) > 0)

        uba = not lba
    else:
        alp = -gTp/pTGp           # unconstrained optimal step
        lba = bool(alp <= alpu)   # lower bound active
        uba = bool(alp >= alpo)   # upper bound active

    if lba: alp = alpu
    if uba: alp = alpo

    # print?
    if abs(alp) == numpy.inf:
        print(gTp, pTGp, alpu, alpo, alp, lba, uba, ier)

    return alp, lba, uba, ier


#-----
def ldldown(L, d, j):
    """
      function L, d = ldldown(L, d, j)

      Downdates LDL^T factorization when j-th row and column are replaced
      by j-th unit vector.

      d contains diag(D) and is assumed positive
    """

    n = len(d)

    test = 0
    if test:
        logger.debug('enter ldldown')
        A = L.dot(diag(d).dot(L.T))
        A[:,j] = numpy.zeros(n)
        A[j] = numpy.zeros((1,n))
        A[j,j] = 1

    if j < n:
        I = numpy.arange(j)
        K = numpy.arange(j+1,n)
        L[K[:,None],K], d[K], p = ldlrk1(L[K[:,None],K], d[K], d[j], L[K,j])
      # update jth row and column
        L[j,I] = numpy.zeros((    j,))
        L[K,j] = numpy.zeros((n-1-j,))
    else:
        L[n-1,:n-2] = numpy.zeros((1,n-2))
    d[j] = 1

    if test:
        A1 = L.dot(diag(d)).dot(L.T)
        quot = numpy.linalg.norm(A1-A,1)/numpy.linalg.norm(A,1)
        logger.debug('leave ldldown')

    return L, d


#-----
def ldlrk1(L, d, alp, u):
    """
      function L, d, p = ldlrk1(L, d, alp, u)

      Computes LDL^T factorization for LDL^T+alp*uu^T if alp>=0 or if the new factorization is
      definite (both signalled by p=[]); otherwise, the original L,d and a direction p of null
      or negative curvature are returned.

      d contains diag(D) and is assumed positive

      Note: does not work for dimension 0.
    """

    test = 0
    if test:
        logger.debug('enter ldlrk1')
        A = L.dot(diag(d).dot(L.T)) + alp*(u.dot(u.T))

    p = numpy.array([])
    if alp == 0:
        return L, d, p

    n = len(u)
    neps = n*numpy.spacing(1)

    # save old factorization
    L0 = copy.copy(L)
    d0 = copy.copy(d)

    # update
    for k in find(u != 0):
        k = int(k)
        del_ = d[k] + alp*u[k]**2
        if alp < 0 and del_ <= neps:
            # update not definite
            p = numpy.zeros(n)
            p[k] = 1
            p[:k] = numpy.linalg.solve(L[:k,:k].T, p[:k])
            # restore original factorization
            L = L0
            d = d0
            if test:
                indef = ((p.T).dot(A.dot(p))) / ((numpy.abs(p).T).dot((numpy.abs(A)).dot(numpy.abs(p))))
                logger.debug('leave ldlrk1 at 1')

            return L, d, p

        q = d[k]/del_
        d[k] = del_
        # in C, the following 3 lines would be done in a single loop
        # WLAVWLAVWLAV
        ind = numpy.arange(k, n)  # TODO: why n-1 instead of n?
#        if len(L.shape) == 1:
#            L = L.reshape(1, len(L))
        c = L[ind,k].dot(u[k])
        L[ind,k] = numpy.outer(L[ind,k], q).flatten() + numpy.outer((alp*(u[k])/del_), u[ind]).flatten()
        u[ind] = u[ind] - c
        alp = alp*q
        if alp == 0:
            break

    if test:
        A1 = L.dot(diag(d).dot(L.T)), A
        quot = numpy.linalg.norm(A1-A,1)/numpy.linalg.norm(A,1)
        logger.debug('leave ldlrk1 at 2')

    return L, d, p


#-----
def ldlup(L, d, j, g):
    """
      function L, d, p = ldlup(L, d, j, g)

      Updates LDL^T factorization when a unit j-th row and column are replaced by column g
      if the new matrix is definite (signalled by p=[]); otherwise, the original L,d and
      a direction p of null or negative curvature are returned

      d contains diag(D) and is assumed positive

      Note that g must have zeros in other unit rows!!!
    """

    p = numpy.array([])

    test = 0
    if test:
        logger.debug('enter ldlup')
        A = L.dot(diag(d).dot(L.T))
        A[:,j] = g
        A[j,:] = g.T

    n = len(d)
    I = numpy.arange(0, j)
    K = numpy.arange(j+1, n)
    del_ = 0
    if j == 0:
        v = numpy.array([])
        del_ = g[j]
        if del_ <= n*numpy.spacing(1):
            p = numpy.concatenate((1, zeros(n-2)))
            if test:
                logger.debug('A = %s', A)
                logger.debug('p = %s', p)
                Nenner = numpy.abs(p).T.dot(numpy.abs(A).dot(numpy.abs(p)))
                if Nenner == 0:
                    indef1 = 0
                else:
                    indef1 = (p.T.dot(A.dot(p)))/Nenner
                logger.debug('indef1 = %s', indef1)
                logger.debug('leave ldlup at 1')

            return L, d, p

        w = g[K]/del_
        L[j, I] = v.T
        d[j] = del_
        if test:
            A1 = L.dot(diag(d)).dot(L.T)
            logger.debug('A1 = %s', A1)
            logger.debug('A = %s', A)
            logger.debug('quot = %s', numpy.linalg.norm(A1-A, 1)/numpy.linalg.norm(A, 1))
            logger.debug('leave ldlup at 3')

        return L, d, p

    try:
        u = numpy.linalg.solve(L[I[:,None],I], g[I])
    except numpy.linalg.LinAlgError as e:
        # happens if len(I) == 1, so simply divide
        u = g[I]/L[I,I]
    v = u/d[I]
    del_ = g[j] - u.T.dot(v)
    if del_ <= n*numpy.spacing(1):
        try:
            p1 = numpy.linalg.solve(L[I[:,None],I].T, v)
        except numpy.linalg.LinAlgError:
            # happens if len(I) == 1, so simply divide
            p1 = v/L[I,I]
        p = numpy.concatenate((p1, numpy.array((-1,)), numpy.zeros(n-j-1)))
        if test:
            logger.debug('A = %s', A)
            logger.debug('p = %s', p)
            logger.debug('indef1 = %s', (p.T.dot(A.dot(p)))/(numpy.abs(p).T.dot(numpy.abs(A).dot(numpy.abs(p)))))
            logger.debug('leave ldlup at 2')

        return L, d, p

    w = (g[K]-L[K[:,None],I].dot(u))/del_
    L[K[:,None],K], d[K], q = ldlrk1(L[K[:,None],K], d[K], -del_, w.copy())
    if q.size <= 0:
        L[j,I] = v.T
        L[K[:,None],j] = w[:,None]
        d[j] = del_
        if test:
            A1 = L.dot(diag(d).dot(L.T))
            logger.debug('A1 = %s', A1)
            logger.debug('A = %s', A)
            logger.debug('quot = %s', numpy.linalg.norm(A1-A,1)/numpy.linalg.norm(A,1))
            logger.debug('leave ldlup at 4')

    else:
        r"""
        % work around expensive sparse L(K,K)=LKK
        L=[L(1:j,:); LKI,L(K,j),LKK];
        pi=w'*q;
        p=[LII'\(pi*v-LKI'*q);-pi;q];
        """
        # TODO: there is something missing in this mapping (extension of matrices), for now skip
        # this update and let upstream deal with it.
        """
        pi_ = numpy.outer(w.T, q)
        p = numpy.concatenate((numpy.linalg.solve(L[I[:,None],I].T, (pi_*v - L[K[:,None],I].T.dot(q))), - pi_, q))
        if test:
            logger.debug('indef2 = %s', (p.T.dot(A.dot(p)))/(numpy.abs(p).T.dot(numpy.abs(A).dot(numpy.abs(p)))))
            logger.debug('leave ldlup at 5')
        """
        logger.debug('DEBUG: skipping update step in ldlup')

    return L, d, p



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function near, d = snobnn(x0, x, m, dx)

 Computes a safeguarded set of m nearest neighbors to a point x0
 for each coordinate, the nearest point differing from x0 in the ith
 coordinate by at least 1 is chosen.
 
 Input:
  x0            point for which the neighbors are to be found
  x             the rows are the points from which the neighbors are to be
                chosen (possibly x0 is among these points)
  m             number of neighbors to be found
  dx            resolution vector
 
 Output:
  near          vector pointing to the nearest neighbors (i.e. to the rows
                of x)
  d             maximal distance between x0 and a point of the set of m
                nearest neighbors
"""

from ._gen_utils import rsort, max_, find, std, sort
import numpy


def snobnn(x0, x, m, dx):
    n = len(x0)         # dimension of the problem
    d = numpy.sqrt(numpy.sum( (x - x0*numpy.ones(x.shape[1]))**2, 1))
    d1, ind = sort(d)

    if not d1[0]:
        ind = ind[1:]	# eliminate x0 if it is in the set

    near = numpy.array([])
    for i in range(n):
        j = numpy.amin( find(numpy.abs(x[ind,i] - x0[i]) - dx[i] >= 0) )
        near = numpy.append(near, ind[j])
        ind = numpy.delete(ind, j)

    j = 0
    while len(near) < m:
        if (not find(near == ind[j]).any()) and (numpy.amax(numpy.abs(x0 - x[ind[j],:]) - dx) >= 0):
            near = numpy.append(near, ind[j])

        j += 1

    near = near.astype(int)
    d, ind = sort(d[near])
    near = near[ind]
    d = numpy.amax(d)

    return near, d



#!/usr/bin/env python

import codecs, os, re, glob
from setuptools import setup, find_packages
from codecs import open


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# https://packaging.python.org/guides/single-sourcing-package-version/
def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='SQSnobFit',
    version=find_version('python', 'SQSnobFit', '_version.py'),
    description='SnobFit - Stable Noisy Optimization by Branch and FIT',
    long_description=long_description,

    url='http://scikit-quant.org/',

    maintainer='Wim Lavrijsen',
    maintainer_email='WLavrijsen@lbl.gov',

    license='other',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development',

        'License :: Other/Proprietary License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        'Natural Language :: English'
    ],

    install_requires=['numpy', 'SQCommon'],

    keywords='quantum computing optimization',

    package_dir={'': 'python'},
    packages=find_packages('python', include=['SQSnobFit']),
)



from __future__ import print_function
import logging

log = logging.getLogger('SKQ.SnobFit')
if not log.hasHandlers():
    def _setupLogger(log):
        import sys
        hdlr = logging.StreamHandler(sys.stdout)
        frmt = logging.Formatter('%(name)-12s: %(levelname)8s %(message)s')
        hdlr.setFormatter(frmt)
        log.addHandler(hdlr)
        log.propagate = False
    _setupLogger(log)
    del _setupLogger


log.info("""
------------------------------------------------------------------------
W. Huyer and A. Neumaier, "Snobfit - Stable Noisy Optimization by Branch and Fit",
 ACM Trans. Math. Software 35 (2008), Article 9.
Software available at www.mat.univie.ac.at/~neum/software/snobfit
------------------------------------------------------------------------""")

__all__ = ['minimize', 'optset', 'log', 'snobfit']

from ._snobfit import minimize, snobfit
from ._optset import optset



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 request, xbest, fbest = snobfit(x, f, config, dx=None)
 minimization of a function over a box in R^n

 Input:
  file         name of file for input and output
               if nargin < 5, the program continues a previous run and
               reads from file.mat the output is (again) stored in file.mat

^^do not use file - store variables globally,
or make them available to be passed in?

  x            the rows are a set of new points entering the
               optimization algorithm together with their function
               values
  f            matrix containing the corresponding function values
               and their uncertainties, i.e., f(j,1) = f(x(j))
               and f(j,2) = df(x(j))
               a value f(j,2)<=0 indicates that the corresponding
               uncertainty is not known, and the program resets it to
               sqrt(numpy.spacing(1))
  config       structure variable defining the box [u,v] in which the
               points are to be generated, the number nreq of
               points to be generated and the probability p that a
               point of type 4 is generated
               config = struct('bounds',{u,v},'nreq',nreq,'p',p)
  dx           only used for the definition of a new problem (when
               the program should continue from the values stored in
               file.mat, the call should have only 4 input parameters!)
               n-vector (n = dimension of the problem) of minimal
               stnp.spacing(1), i.e., two points are considered to be different
               if they differ by at least dx(i) in at least one
               coordinate i

 Output:
  request      nreq x (n+3)-matrix
               request[j,1:n] is the jth newly generated point,
               request[j,n+1] is its estimated function value and
               request[j,n+3] indicates for which reason the point
               request[j,1:n] has been generated
               request[j,n+3] = 1 best prediction
                              = 2 putative local minimizer
                              = 3 alternative good point
                              = 4 explore empty region
                              = 5 to fill up the required number of
                              function values if too little points of
                              the other classes are found
  xbest        current best point
  fbest        current best function value (i.e. function value at xbest)
"""

from SQCommon import Result, ObjectiveFunction
from ._gen_utils import diag, max_, min_, find, extend, rand, sort
from ._optset    import optset
from ._snobinput import snobinput
from ._snoblocf  import snoblocf, snobround
from ._snoblp    import snoblp
from ._snobnan   import snobnan
from ._snobpoint import snobpoint
from ._snobqfit  import snobqfit
from ._snobsplit import snobsplit
from ._snobupdt  import snobupdt
from ._snob5     import snob5
import logging
import pickle
import math
import numpy

__all__ = ['minimize', 'log']

log = logging.getLogger('SKQ.SnobFit')


_im_storage = None
def _snobsave(filename, *args):
    if filename is None:
        global _im_storage
        _im_storage = args
    else:
        pickle.dump(args, open(filename, 'wb'))

def _snobload(filename):
    if filename is None:
        global _im_storage
        return _im_storage
    return pickle.load(open(filename, 'rb'))


'''
iter - number of iterations to run snobfit loop
f - use numpy array
x - use numpy multidimensional array
config - use dictionary
    bounds - multidimensional numpy array
    nreq - int
    probability p - float
'''
def snobfit_func(iter, x, func, config, dx, df):
    req = x
    f = numpy.zeros((len(x), 2))

    for n in iter:
        for ind in len(req):
            f[ind][0] = func(req[ind])
            f[ind][1] = df

        req, xbest, fbest = snobfit(req, f, config, dx)

    return req, xbest, fbest


def fill_request(request, func, nparams):
    x = numpy.zeros((len(request), nparams))
    f = numpy.zeros((len(request), 2))
    for i in range(len(request)):
        x[i] = request[i][0:nparams]
        try:
            res = func(x[i])
            err = math.sqrt(numpy.spacing(1))
            if type(res) == tuple:
                if res[1] != 0:
                    err = res[1]
                res = res[0]
            f[i] = (res, err)
        except Exception as e:
            log.Error('Function evaluation failed: %s', str(e))
            f[i] = numpy.nan
    return x, f


#-----
def minimize(f, x0, bounds, budget, optin={}, **optkwds):
  # The user-facing API is the equivalent of snobdriver, providing the loop
  # over the "internal" snobfit function.
    if budget <= 0:
        budget = 100000

    if type(x0) != numpy.ndarray:
        x0 = numpy.array(x0)

    if type(bounds) != numpy.ndarray:
        bounds = numpy.array(bounds)

    objfunc = ObjectiveFunction(f, options={'simple_function' : True})

  # calculate resolution vector from the bounds
    dx = (bounds[:,1]-bounds[:,0])*1E-5

  # setup parameters (TODO: use optin/optkwds)
    if type(optin) == dict:
        options = optset(**dict(optin, **optkwds))
    else:
        options = optset(optin, **optkwds)

  # minimum number of function values before considering stopping
    minfcall = options.minfcall
    if minfcall is None:
        minfcall = len(bounds)*5

    config = {"bounds": bounds, "nreq": 2*len(bounds)+6, "p": .5}
    if optin is not None:
        if options.maxmp is not None and 0 < options.maxmp:
            config["nreq"] = options.maxmp

    nstop = options.maxfail      # number of times no improvement is tolerated

    nparams = len(bounds)

    if x0 is None or (type(x0) == numpy.ndarray and not x0.shape) or not len(x0):
      # initial call with empty list
        request, xbest, fbest = snobfit(numpy.array([]).reshape(0, len(bounds)),
            numpy.array([]).reshape(0, 2),
            config, dx)
    else:
      # initial call with just x0 as input point(s) (establishes initial request)
        request, xbest, fbest = snobfit(x0.reshape(1, len(x0)),
            numpy.array([objfunc(x0), math.sqrt(numpy.spacing(1))]).reshape(1, 2),
            config, dx)

    if options.verbose:
        print('request =', request)

  # calculate the requested points and set uncertainties
    x, vals = fill_request(request, objfunc, nparams)

    ncall0 = len(vals)                 # initial budget used
    fbestn, jbest = min_(vals[:,0])    # best function value
    if fbestn < fbest:
        fbest = fbestn
        xbest = x[jbest,:]

  # display current number of function values, best point and function value
    log.info('# calls = %d; xbest = %s; fbest = %f', ncall0, str(xbest), fbest)

    nstop0 = 0;
  # repeated calls to Snobfit
    while ncall0 < budget:   # repeat till ncall function values are reached
                             # (if the stopping criterion is not fulfilled first)
        request, xbest, fbest = snobfit(x, vals, config)
        if options.verbose:
            print('request =', request)

      # computation of the function values at the suggested points
        x, vals = fill_request(request, objfunc, nparams)

      # update function call counter
        ncall0 = ncall0 + len(vals)
        fbestn, jbest = min_(vals[:,0])    # best function value
        if fbestn < fbest:
            fbest = fbestn
            xbest = x[jbest,:]

          # display current number of function values
            log.info('# calls = %d; xbest = %s; fbest = %f', ncall0, str(xbest), fbest)

            nstop0 = 0
        elif budget >= minfcall:
            nstop0 = nstop0 + 1

      # check stopping criterion
        if nstop0 >= nstop and ncall0 >= minfcall:
            break

    return Result(fbest, xbest), objfunc.get_history()


def snobfit(x, f, config, dx = None):
    ind = find(f[:,1] <= 0)
    if not (ind.size <= 0 or numpy.all(ind==0)):
        f[ind,1] = math.sqrt(numpy.spacing(1))    # may be wrong

    rho = 0.5*(math.sqrt(5)-1)	  # golden section number
    bounds = config['bounds']
    u1 = bounds[:,0].reshape(1, len(bounds))    # lower
    v1 = bounds[:,1].reshape(1, len(bounds))    # upper

    nreq = config['nreq']
    p = config['p']
    n = u1.shape[1]     # dimension of the problem
    nneigh = n+5        # number of nearest neighbors

    dy = 0.1*(v1-u1)    # defines the vector of minimal distances between two
                        # points suggested in a single call to Snobfit

    filename = config.get('filename', None)

    if dx is not None:  # a new job is started
        if numpy.any(dx<=0):
            raise ValueError('dx should contain only positive entries')

        if dx.shape[0] > 1:
            dx = dx.T

        if x.size > 0:
            u = numpy.minimum(x.min(axis=0), u1)
            v = numpy.maximum(x.max(axis=0), v1)
        else:
            u = u1.copy()
            v = v1.copy()

        x, f, np, t = snobinput(x, f)   # throw out duplicates among the points
                                        # and compute mean function value and
                                        # deviation
        if x.size > 0:
            xl, xu, x, f, nsplit, small = snobsplit(x, f, u, v, None, u, v)
            d = numpy.inf*numpy.ones((1, len(x)))
        else:
            xl = numpy.array([])
            xu = numpy.array([])
            nsplit = numpy.array([])
            small = numpy.array([])

        notnan = find(numpy.isfinite(f[:,0]))
        if notnan.size > 0:
            fmn = min_(f[notnan,1])
            fmx = max_(f[notnan,1])
        else:
            fmn = 1
            fmx = 0

        if len(x) >= nneigh+1 and fmn < fmx:
            inew = range(len(x))
            near = numpy.zeros((len(x), nneigh))
            d = numpy.zeros(len(x))
            for j in inew:
                near[j], d[j] = snobnn(x[j], x, nneigh, dx)

            fnan = find(numpy.isnan(f[:,0]))
            if fnan.size > 0:
                f = snobnan(fnan, f, near, inew)

            jsize = inew[-1]
            y = numpy.zeros((jsize, 2))
            g = numpy.zeros((jsize, 2))
            sigma = numpy.zeros(jsize)
            f = extend(f, 1)
            for j in inew:
                y[j], f[j, 2], c, sigma[j] = snoblocf(j, x, f[:,0:2], near, dx, u, v)
                g[j] = c.reshape(1, len(c))

            fbest, jbest = min_(f[:,0])
            xbest = x[jbest]
        else:
            fnan = numpy.array([], dtype=int)
            near = numpy.array([], dtype=int)
            d = numpy.inf*numpy.ones((1,len(x)))

            x5 = snob5(x, u1, v1, dx, nreq)
            request = numpy.concatenate((x5, numpy.nan*numpy.ones((nreq,1)), 5*numpy.ones((nreq,1))), 1)
            if x.size > 0 and f.size > 0:
                fbest, jbest = min_(f[:,0])
                xbest = x[jbest]
            else:
                xbest = numpy.nan*numpy.ones((1,n))
                fbest = numpy.inf

            if len(request) < nreq:
                snobwarn()

            y = None
            _snobsave(filename, xbest, fbest, x, f, xl, xu, y, nsplit, small, near, d, np, t, fnan, u, v, dx)
            return request, xbest, fbest
    else:
        xnew = x.copy()
        fnew = f.copy()
        xbest, fbest, x, f, xl, xu, y, nsplit, small, near, d, np, t, fnan, u, v, dx = _snobload(filename)

        nx = len(xnew)
        oldxbest = xbest

        xl, xu, x, f, nsplit, small, near, d, np, t, inew, fnan, u, v  = \
            snobupdt(xl, xu, x, f, nsplit, small, near, d, np, t, xnew, fnew, fnan, u, v, u1, v1, dx)

        if near.size > 0:
            ind = find(numpy.isnan(f[:,0]))
            if ind.size > 0:
                fnan = numpy.concatenate((fnan, ind.flatten()))
            if fnan.size > 0:
                f = snobnan(fnan, f, near, inew)

            fbest, jbest = min_(f[:,0])
            xbest = x[jbest]
            jsize = int(inew[-1]+1)
            if y is None:
                y = numpy.zeros((jsize, x.shape[1]))
            else:
                y = numpy.append(y, numpy.zeros((jsize-len(y), x.shape[1])), axis=0)
            g = numpy.zeros((jsize, x.shape[1]))
            sigma = numpy.zeros(jsize)
            if f.shape[1] < 3:
               f = extend(f, 3-f.shape[1])

            for j in inew:
                y[j], f[j,2], c, sigma[j] = snoblocf(j, x, f[:,0:2], near, dx, u, v)
                g[j] = c

        else:
            x5 = snob5(x, u1, v1, dx, nreq)
            request = numpy.concatenate((x5, numpy.NaN*numpy.ones((nreq,1)), 5*numpy.ones((nreq,1))), 1)
            if x.size > 0:
                (fbest, ibest) = min_(f[:,0])
                xbest = x[ibest]
            else:
                xbest = numpy.array([])
                fbest = numpy.inf

            if request.shape[0] < nreq:
                snobwarn()

            _snobsave(filename, xbest, fbest, x, f, xl, xu, y, nsplit, small, near, d, np, t, fnan, u, v, dx)
            return request, xbest, fbest

    sx = len(x)
    request = numpy.array([]).reshape(0, x.shape[1]+2)
    ind = find(numpy.sum(numpy.logical_and(xl <= numpy.outer(numpy.ones(sx), v1), xu >= numpy.outer(numpy.ones(sx), u1)), 1) == n)
    minsmall, k = min_(small[ind])
    maxsmall = small[ind].max(0)
    m1 = numpy.floor((maxsmall-minsmall)/3)
    k = find(small[ind] == minsmall)
    k = ind[k].flatten()
    fsort, j = sort(f[k,0])
    k = k[j]
    isplit = k[0]

    if numpy.sum(numpy.logical_and(u1<=xbest, xbest<=v1)) == n:
        z, f1 = snobqfit(jbest, x, f[:,0], near, dx, u1, v1)
    else:
        fbes, jbes = min_(f[ind,0])
        jbes = ind[jbes]
        xbes = x[jbes]
        z, f1 = snobqfit(jbes, x, f[:,0], near, dx, u1, v1)

    z = snobround(z, u1, v1, dx)
    zz = numpy.outer(numpy.ones(sx), z)
    j = find(numpy.sum(numpy.logical_and(xl<=zz, zz<=xu), 1) == n)
    if len(j) > 1:
        msmall, j1 = min_(small[j])
        j = j[j1]

    if numpy.min(numpy.max(numpy.abs(x - numpy.outer(numpy.ones(sx), z)) - numpy.outer(numpy.ones(sx), dx))) >= -numpy.spacing(1):
        dmax = numpy.max((xu[j] - xl[j])/(v-u))
        dmin = numpy.min((xu[j] - xl[j])/(v-u))
        if dmin <= 0.05*dmax:
            isplit = numpy.append(isplit, j)
        else:
            request = numpy.vstack((request, numpy.concatenate((z, numpy.array((f1, 1), ndmin=2)), axis=1)))

    if len(request) < nreq:
        globloc = nreq - len(request)
        glob1 = globloc*p
        glob2 = math.floor(glob1)
        if rand(1) < glob1 - glob2:
            glob = glob2 + 1
        else:
            glob = glob2

        loc = globloc - glob
        if loc:
            local, nlocal = snoblp(f[:,0], near, ind)
            fsort, k = sort(f[local,2]) #uhhhhhh
            j = 0
            sreq = len(request)
            while sreq < (nreq-glob) and j < len(local):
                l0 = local[k[j]]
                y1 = snobround(y[l0], u1, v1, dx)
                yy = numpy.outer(numpy.ones((len(x), 1)), y1)
                l = find(numpy.sum(numpy.logical_and(xl<=yy, yy<=xu), 1) == n)
                if len(l) > 1:
                    msmall, j1 = min_(small[l])
                    l = l[j1]

                dmax = numpy.max((xu[l] - xl[l]) / (v - u))
                dmin = numpy.min((xu[l] - xl[l]) / (v - u))
                if dmin <= 0.05*dmax:
                    isplit = numpy.append(isplit, l)
                    j += 1
                    continue

                if numpy.max(abs(y1-x[l])-dx) >= -numpy.spacing(1) and \
                        (not sreq or numpy.min( \
                             numpy.max(numpy.abs(request[:,0:n] - \
                             numpy.outer(numpy.ones(sreq), y1))-numpy.outer(numpy.ones(sreq), numpy.maximum(dy,dx)), axis=1)) >= -numpy.spacing(1)):
                    if numpy.sum(y1 == y[l0]) < n:
                        D = f[l0,1]/dx**2
                        #Possibly problem area!
                        f1 = f[l0,0] + g[l0].dot((y1 - x[l0]).T) + sigma[l0]*( \
                            (y1 - x[l0]).dot(diag(D).dot((y1-x[l0]).T) + f[l0,1]))
                    else:
                        f1 = f[l0,2]
                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 2), ndmin=2)), axis=1)))

                sreq = len(request)
                j += 1

            if sreq < nreq-glob:
                fsort, k = sort(f[nlocal,2])

            j = 0
            while sreq < (nreq-glob) and j < len(nlocal):
                l0 = nlocal[k[j]]
                y1 = snobround(y[l0], u1, v1, dx)
                yy = numpy.outer(numpy.ones(len(x)), y1)
                l = find(numpy.sum(numpy.logical_and(xl<=yy, yy<=xu),1) == n)
                if 1 < len(l):
                    l = l[min_(small[l])[1]]

                j += 1       # early in body, b/c of continue on error

                normdiff = (xu[l] - xl[l]) / (v - u)
                if not len(normdiff):
                    continue

                dmax = numpy.max(normdiff)
                dmin = numpy.min(normdiff)
                if dmin <= 0.05*dmax:
                    isplit = numpy.append(isplit, l)
                    continue

                if numpy.max(numpy.abs(y1-x[l]) - dx) >= -numpy.spacing(1) and \
                        (not sreq or numpy.min( \
                             numpy.max(numpy.abs(request[:,:n] - numpy.outer(numpy.ones(sreq), y1)) - \
                             numpy.outer(numpy.ones(sreq), numpy.maximum(dy,dx)), axis=1)) >= -numpy.spacing(1)):
                    if numpy.sum(y1==y[l0]) < n:
                        D = f[l0,1]/(dx**2)
                        f1 = f[l0,0] + g[l0].dot((y1 - x[l0]).T) + \
                        sigma[l0]*(((y1 - x[l0]).dot(diag(D).dot((y1 - x[l0]).T))) + f[l0,1])
                    else:
                        f1 = f[l0,2]

                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 3), ndmin=2)), axis=1)))

                sreq = len(request)


    sreq = len(request)
    for l in isplit.flatten():
        if sreq == nreq:
            break

        jj = find(ind==l)
        ind = numpy.delete(ind, jj) #ind(jj) = []
        y1, f1 = snobpoint(x[l], xl[l], xu[l], f[l,0:2], g[l], sigma[l], u1, v1, dx)

        if numpy.max(numpy.abs(y1-x[l]) - dx) >= -numpy.spacing(1) and \
                (not sreq or numpy.min( \
                     numpy.max(numpy.abs(request[:,:n] - numpy.outer(numpy.ones(sreq), y1)) - \
                     numpy.outer(numpy.ones(sreq), dx), axis=1)) >= -numpy.spacing(1)):
            request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4), ndmin=2)), axis=1)))
        sreq = len(request)

    first = True
    while (sreq < nreq) and ind.size > 0:   # and find(small[ind] <= (minsmall + m1)).any():
        try:
            im1 = int(m1+1)
        except OverflowError:     # happens with m1 == inf, b/c already converged
            break

        for m in range(int(m1+1)):
            if first:
                first = False
                continue

            m = 0
            k = find(small[ind] == minsmall+m)
            while k.size <= 0:
                m += 1
                k = find(small[ind] == minsmall+m)

            if k.size > 0:
                k = ind[k].flatten()
                fsort, j = sort(f[k,0])
                k = k[j]
                l = int(k[0])
                jj = find(ind == l)
                ind = numpy.delete(ind, jj)
                y1, f1 = snobpoint(x[l], xl[l], xu[l], f[l,0:2], g[l], sigma[l], u1, v1, dx)
                if numpy.max(numpy.abs(y1-x[l]) - dx) >= -numpy.spacing(1) and \
                        (not sreq or numpy.min( \
                             numpy.max(numpy.abs(request[:,:n] - numpy.outer(numpy.ones(sreq), y1)) - \
                             numpy.outer(numpy.ones(sreq), numpy.maximum(dy,dx)), axis=1)) >= -numpy.spacing(1)):
                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4), ndmin=2)), axis=1)))

                sreq = len(request)
                if sreq == nreq:
                    break
            m = 0

    if len(request) < nreq:
        x5 = snob5(numpy.concatenate((x, request[:,:n])), u1, v1, dx, nreq - len(request))
        nx = len(x)
        for j in range(len(x5)):
            x5j = x5[(j,),:]
            i = find(numpy.sum(numpy.logical_and( \
                         xl <= numpy.outer(numpy.ones(nx), x5j), numpy.outer(numpy.ones((nx,1)), x5j) <= xu),
                     1) == n)
            if i < len(i):
                i = i[min_(small[i])[1]]

            D = f[i,1]/(dx**2)
            f1 = f[i,0] + (x5j - x[i])*g[i].T + sigma[i]*(x5j - x[i])*(diag(D)*(x5j - x[i]).T) + f[i,1]
            request = numpy.vstack((request, numpy.concatenate((x5j, numpy.array((f1, 5), ndmin=2)), axis=1)))

    if len(request) < nreq:
        snobwarn()

    _snobsave(filename, xbest, fbest, x, f, xl, xu, y, nsplit, small, near, d, np, t, fnan, u, v, dx)
    return request, xbest, fbest


def snobwarn():
    """issues a warning if SNOBFIT has not been able to generate the desired
       number of points"""

    import warnings
    warnings.warn("WARNING: The algorithm was not able to generate the desired number of points\n"
                  "Change the search region or refine dx")
